// Libregresion -- A library to help you with
// your linear regresions, interpolations, systems solving, etc...
//
// (c) 2003, Jaime Anguiano Olarra <jaime@gnome.org>
//	Fabian Seoane <fseoane@wanadoo.es>
//


using System;


namespace Libregresion{
	

public class LinearAlgebra{	
	

	public static double[,] LUP( double[,] A, int n, ref int[] P, ref bool sign)
	{/* Factorise a general n x n real matrix A into,
	*
	*   P A = L U
	*
	* where P is a permutation matrix, L is unit lower triangular and U
	* is upper triangular.
	*
	* L is stored in the strict lower triangular part of the input
	* matrix. The diagonal elements of L are unity and are not stored.
	*
	* U is stored in the diagonal and upper triangular part of the
	* input matrix.
	*
	* P is stored in the array P.
	*
	*  It is equivalent to a Gaussian Elimination and runs in O(n^3) time
	*/

	int i, j, k;
		
	for( i=0; i<n; i++)
		P[i] = i;
	
	for(j=0; j<n-1; j++){
		/* find maximum int the j-th row */
		
		double ajj, max = System.Math.Abs(A[j, j]);
		int i_pivot = j;
		
		for( i = j+1; i<n; i++)
		{
			double aij = System.Math.Abs( A[i,j] );
			
			if( aij > max )
			{
				max = aij;
				i_pivot = i;
			}
		}

		if( i_pivot != j)
		{
			Miscel.SwapRows( ref A, j, i_pivot);
			Miscel.SwapPermutation( ref P, j, i_pivot);
			sign = !(sign);
		}
		
		ajj = A[j,j];
		
		if( ajj == 0.0 )
			throw new LibregresionException("Singular Matrix ecountered");
		
		else
		{
			for( i=j+1; i<n; i++)
			{
				double aij = A[i,j]/ajj;
				A[i,j] = aij;
				
				for( k=j+1; k<n; k++)
				{
					double aik = A[i,k];
					double ajk = A[j,k];
					A[i,k] = aik - aij*ajk;
				}
			}
		}
	}
	return A;
}
		
		

	public static double Det( double[,] A)
	{	bool sign = true;
		//calculates the determinant by performing an LUP-decomposition
		try{
		int n = A.GetLength(0);
		double det=1;
		if(n != A.GetLength(1))
			throw new LibregresionException(" Not Square Matrix ");
		if( n==1 )
			return A[0,0];
		if( n==2 )
			return (A[0,0]*A[1,1] - A[1,0]*A[0,1] );	//determinantes sencillos

		int[] P = new int[n];
		double[,] LU = new double[n,n];
		LU = LUP(A, n, ref P, ref sign);
		for(int i=0; i<n; i++)
			det = det*LU[i,i];
		if( sign )
			return det;
		return -det;

		} catch( SingularMatrixException){
		return 0;
		}
	}


	public static double[] Solve( double[,] A, double[] b)
	{

		int n = A.GetLength(0);
		int[] P = new int[n];
		bool sign = true;
		
		if( n != A.GetLength(1))
			throw new LibregresionException(" Not Square Matrix");
		
		A = LUP(A, n,ref P, ref sign);
		
		double[] y = new double[n];
		double[] x = new double[n];
		double sum;
		int i, j;
		
		/* Solve Ly == Pb */		
		y[0] = b[P[0]];
		for( i = 1; i<n; i++)
		{
			y[i] = b[P[i]];
			for( j= 0; j<i; j++)
			{
				y[i] -= A[i,j]*y[j];
			}
		}
		
		
		/* Solve Ux = y*/
		x[n-1] = y[n-1]/(A[n-1, n-1]);
		for( i = n-2; i>=0; i--)
		{
			x[i] = y[i];
			for( j = i+1; j<=n-1; j++)
			{
				x[i] -= A[i,j]*x[j];
			}
			x[i] = x[i]/A[i,i];
		}
		
		return x;
		
		
		
	}

	public static double[,] Inverse( double[,] A )
	{
		
		int n = A.GetLength(0);
		if( n != A.GetLength(1))
			throw new LibregresionException(" Not Square Matrix");
		
		double[] I;
		double[] temp = new double[n];
		double[,] inv = new double[n,n];
		I = new double[n];
		for( int i = 0; i<n; i++)
		{
			
			I[i] = 1;
			temp = Solve((double[,]) A.Clone(), I);	//hay otra forma para hacer que se tomen arrays por valor??
			//copia a la inversa
			for( int j = 0; j<n; j++)
			{
				inv[j,i] = temp[j];
			}
			I[i] = 0;
		}
		
		return inv;
	}

	


	
	
	/* ******************************** 
	*
	*   Ahora lo mismo para complejos 
	*
	**********************************/
	
	
	static Complex[,] LUP( Complex[,] A, int n, ref int[] P, ref bool sign)
	{
	
	int i, j, k;
		
	for( i=0; i<n; i++)
		P[i] = i;
	
	for(j=0; j<n-1; j++){
		/* find maximum int the j-th row */
		
		double max = A[j, j].Norm;
		Complex ajj;
		int i_pivot = j;
		
		for( i = j+1; i<n; i++)
		{
			double aij =  A[i,j].Norm;
			
			if( aij > max )
			{
				max = aij;
				i_pivot = i;
			}
		}
		
		if( i_pivot != j)
		{
			Miscel.SwapRows( ref A, j, i_pivot);
			Miscel.SwapPermutation( ref P, j, i_pivot);
			sign = !(sign);
		}
		
		ajj = A[j,j];
		
		if( ajj.Re == 0 && ajj.Im == 0 )
			throw new LibregresionException( " Singular Matrix Encountered during LU-decomposition");
			
		else
		{
			for( i=j+1; i<n; i++)
			{
				Complex aij = A[i,j]/ajj;
				A[i,j] = aij;
				
				for( k=j+1; k<n; k++)
				{
					Complex aik = A[i,k];
					Complex ajk = A[j,k];
					A[i,k] = aik - aij*ajk;
				}
			}
		}
	}
	
	return A;
}


	public static Complex[] Solve( Complex[,] A, Complex[] b)
	{

		int n = A.GetLength(0);
		int[] P = new int[n];
		bool sign = true;
		
		if( n != A.GetLength(1))
			throw new LibregresionException(" Not Square Matrix");
		
		A = LUP(A, n,ref P, ref sign);
		
		Complex[] y = new Complex[n];
		Complex[] x = new Complex[n];
		Complex sum;
		int i, j;
		
		/* Solve Ly == Pb */		
		y[0] = b[P[0]];
		for( i = 1; i<n; i++)
		{
			y[i] = b[P[i]];
			for( j= 0; j<i; j++)
			{
				y[i] -= A[i,j]*y[j];
			}
		}
		
		
		/* Solve Ux = y*/
		x[n-1] = y[n-1]/(A[n-1, n-1]);
		for( i = n-2; i>=0; i--)
		{
			x[i] = y[i];
			for( j = i+1; j<=n-1; j++)
			{
				x[i] -= A[i,j]*x[j];
			}
			x[i] = x[i]/A[i,i];
		}
		
		return x;
	}
		

	public static Complex[,] Inverse( Complex[,] A )
	{
		
		int n = A.GetLength(0);
		if( n != A.GetLength(1))
			throw new LibregresionException(" Not Square Matrix");
		
		Complex[] I;
		Complex[] temp = new Complex[n];
		Complex[,] inv = new Complex[n,n];
		I = new Complex[n];
		for( int i = 0; i<n; i++)
		{
			for(int j=0; j<n; j++){	
				if( j==i)
					I[i] = new Complex(1,0);
				else
					I[j] = new Complex(0,0);
			}
			temp = Solve((Complex[,]) A.Clone(), I);	//hay otra forma para hacer que se tomen arrays por valor??
			//copia a la inversa
			for( int j = 0; j<n; j++)
			{
				inv[j,i] = temp[j];
			}
			I[i] = new Complex(0,0);
		}
		
		return inv;
	}
		
	



	public static Complex Det( Complex[,] A )
	{
		bool sign = true;
		//calculates the determinant by performing an LUP-decomposition
		try{
		int n = A.GetLength(0);
		Complex det=new Complex(1,0);
		if(n != A.GetLength(1))
			throw new LibregresionException(" Not Square Matrix ");
		if( n==1 )
			return A[0,0];	//de verdad necesitas un ordenador para hacer esto?? xDD		
		if( n==2 )
			return (A[0,0]*A[1,1] - A[1,0]*A[0,1] );
	
		int[] P = new int[n];
		Complex[,] LU = new Complex[n,n];
		LU = LUP(A, n, ref P, ref sign);
		for(int i=0; i<n; i++)
			det = det*LU[i,i];
		if( sign )
			return det;
		return -det;
	
		} catch( SingularMatrixException){
		return new Complex(0,0);
	}
		
		
	}




}
}

