//
// Libregresion -- A library to help you with
// your linear regresions, interpolations, systems solving, etc...
//
// (c) 2003-2004, Jaime Anguiano Olarra <jaime@gnome.org>
//	     Fabian Seoane <f.seoane@unix.net>

using System;
using System.IO;
using System.Collections;

namespace Libregresion{

public class Regresion {
	static double[] xarr; // x data 
	static double[] yarr; // y data
	static double[] x_sums; // the x^j sums
	static double y_sum; // the y sum
	static double[] xy_sums; // the y*x^j sums
	static double[,] coeffs; // coeffs matrix for the linear systems methods
	static int order; // order of the polinomy for the regresions
	static bool saved = false; // data saved?
	static readonly double PI = 3.141592653589;
	
	// Do a simple linear regresion
	public static double[] Regres (double[] xarr, double[] yarr)
	{
		order = 1;
		return Regres (xarr, yarr, 1, 15000);
	}

	// Libregresion for a user choosen number of iterations and order of the polinomy
	public static double[] Regres (double[] xarr, double[] yarr, int n, long iter)
	{
		order = n;
		if (xarr.Length != yarr.Length)
			throw new Exception ();

		// We do the sums for x^j, y and y*x^j
		x_sums = new double[2*n];
		y_sum = Sum(yarr);
		xy_sums = new double[n];

		for (int i=0; i < 2*n; i++) 
			x_sums[i] = Sum(ToNth(xarr, i));
			
		for (int j=0; j < n; j++) 
			xy_sums[j] = Sum (MatrixAlgebra.Multiply (ToNth(xarr, j), yarr));

		// We set the coefficients matrix for the linear system obtained by
		// the regresion method
		coeffs = new double[order, order];
		
		for (int i=0; i < order; i++) {
			for (int j=0; j < order; j++) {
				coeffs[i,j] = x_sums[i+j];
			}
		}

		return LinearAlgebra.Solve( coeffs, xy_sums );
	
	}

	// Returns the sum of all the elements of the array passed to it
	public static double Sum (double[] arr)
	{
		double sum = 0.0;
		
	        for (int i=0; i < arr.Length; i++) {
			sum += arr[i];
		}
		return sum;
	}

	// This method takes the array passed and it powers each of its
	// elements to the power passed, n, (counting from 1). Then it
	// returns a new array with the new values.
	// F.ex. if you pass [2,3] you get [4,9].
	public static double[] ToNth (double[] arr, int n)
	{
		double[] elems = new double[arr.Length];
		double pwr;
		for (int i=0; i < arr.Length; i++) {
			pwr = 1;
			for (int j=0; j < n; j++) {
				pwr *= arr[i];
			}
			elems[i] = pwr;
		}

		return elems;
	}

			

	// Returns the matrix passed multiplied by the constant passed
	public static double[,] Multiply (double[,] x, double c, uint order)
	{
		for (uint i=0; i < order; ++i)
			for (uint j=0; j < order; ++j)
				x[i,j] *= c;
		return x;
	}

		
	// Returns the adjoint of the element given by (n,m) in the matrix x
	public static double[,] AdjointMatrix (double[,] x, int n, int m, uint order)
	{
		double[,] adj;
		uint r = 0, c = 0;
			
		
		adj = new double[order - 1, order - 1];
		for (uint i=0; i < order; ++i) {
			if (i == n)
				continue;
			for (uint j=0; j < order; ++j) {
				if (c == (order - 1)) { 
					++r;
					c = 0;
				}
				if (j == m) 
					continue;
				adj[r,c] = x[i,j];
				++c;
			}
		}
		return adj;
	}
	

	// Returns the determinant of an adjoint matrix element
		
			
	// Returns the sign of the integer passed
	public static int Sign (int s)
	{
		int result;
		result = ((s%2) == 0) ? 1 : -1;
		return result;
	}

	// Returns a copy of the matrix passed
	public static double[,] Copy (double[,] x, uint order)
	{
		return ( double[,] ) x.Clone() ;
	}
	
	

	// Converts an array to a single string form
	public static string ArrToString (double[] arr)
	{
		string s = null;
		for (int i=0; i < arr.Length; ++i)
			s += (arr[i]).ToString() + " ";
		return s;
	}

	

	public static string SolsToString (double[] arr)
	{
		string s = null;
		for (int i=0; i < arr.Length; ++i)
			s += "x" + (i).ToString() + ": " + (arr[i]).ToString() + "  ";
		return s;
	}
	
	// Resolvs a linear system by Gauss-Seidel's method given the
	// coefficient matrix for the unknown variables, a solution vector
	// and the number of iterations it should use.
	public static double[] GaussSeidel (double[,] matrix, double[] solVctr, long iter)
	{
		double q = 0;
		double[] sols = new double[solVctr.Length];

		for (int t=0; t < sols.Length; t++)
			sols[t] = 0;

		for (int k=0; k < iter; k++) {
			for (int i=0; i < solVctr.Length; i++) {
				for (int j=0; j < solVctr.Length; j++) {
					if (i != j) 
						q += matrix[i,j]*sols[j];
				}
				sols[i] = (solVctr[i] - q)/matrix[i,i];
				q = 0;
			}
		}
		return sols;
	} 


	// Check for singularity of the inverse of matrix
	public static bool HasInverse (double[,] matrix, uint n)
	{
		return HasInverse (matrix, n, 0);
	}
		
	// Check for singularity of the inverse of matrix for minimum value epsil
	public static bool HasInverse (double[,] matrix, uint n, double epsil)
	{
		bool response = true;
		for (uint i=0; i < n; ++i) 
			response = (Math.Abs(matrix[i,i]) < epsil) ? false : true;
		return response;
	}	//¿qué se supone que está haciendo??
	
	// Return a unitary matrix of order n
	public static double[,] Unitary (uint n)
	{
		double[,] m;
		m = new double[n,n];

		for (int i=0; i < n; ++i)
			for (int j=0; j < n; ++j) {
				if( i == j )
					m[i,i] = 1; 
			}
		return m;
	}

	// Returns the 'y' interpolated value for the table given by the arrays
	// passed, as well as the iteration and the error
	public static double[] Interpolation (double[] xarr, double[] yarr, double xval)
	{
		int m; // m is the number of intervals (n of points - 1)
		double y, er = 0;
		double n = xarr.Length - 1; // Grade of the interpolation polinomy
		double[] l = new double[xarr.Length];

		m = (int) n;
		for (int i=0; i < n; ++i) {
			y = 1;
			for (int j=0; j < n; ++j) {
				if ((i != j) && ((xarr[j] - xarr[i]) != 0)) {
					y *= ((xval - xarr[j]) / (xarr[i] - xarr[j]));
				}
			}
			l[i] = y;
		}
		y = 0;
		for (int i=0; i < n; ++i)
			y += yarr[i]*l[i];
		
		return new double[3] {y, m, er};
	}

	// Get the factorial of the argument
	public static long Factorial (uint n)
	{
		if (n == 1 || n == 0)
			return 1;

		return (n * Factorial(n - 1));
	}

        // Calculates the binomial coefficient (combinatory n,m)
	public static double Combination (uint n, uint m)
	{
		return (Factorial(n)/(Factorial(m)*Factorial(n - m)));
	}

	// Calculates the variation of n elements taken by m
	public static long Variation (uint n, uint m)
	{
		return (Factorial(n)/Factorial(n - m));
	}
	// ¿ donde se utiliza esto ?
	
	//*** WARNING ***
	//**
	//** Look for a better, general definition of Gamma if any
	// Calculates the gamma of (2*n - 1) , n => 1
	public static double Gamma (uint n)
	{
		double g;
		g = 1;
		for (uint i=1; i <= n; ++i)
			g *= (2*i - 1);
		return (Math.Sqrt(PI)*(1/Math.Pow(2,n))*g);
	}

	// Calcultates the beta of n, m
	public static double Beta (uint n, uint m)
	{
		return (Gamma(n)*Gamma(m)/Gamma(n + m));
	}
	
	// Launches a LibregresionException instance with the message passed
	public static void ThrowException (string msg)
	{
		throw new LibregresionException (msg);
	}

	// Launches a LibregresionException instance with the message passed and the matrix that caused it
	public static void ThrowException (string msg, double[,] x)
	{
		throw new LibregresionException (msg, x);
	}
	
	


}


}
